/**
 * DashbotController.js
 * 
 * $Id$
 * 
 * @package		Dashbot
 * @version		0.2.3
 * @author		Masayuki Iwai <miyabi@mybdesign.com>
 * @copyright	Copyright &copy; 2010 Masayuki Iwai all rights reserved.
 **/


window.DashbotController = function()
{
	var self = this;
	
	var config = {
		"version": "ver.0.2.3", 
		"collapsed_size": {"width": 141, "height": 180}, 
		"expanded_size": {"width": 376, "height": 180}, 
		"back_size": {"width": 376, "height": 344}, 
		"side_margin": 31, 
		"log_scroll_area_width": 203, 
		"log_scrollbar_width": 19, 
		"main_loop_interval": 10, 
		"verification_interval": 360, 
		"tweet_intervals": [1, 30, 60, 120, 180, 360, 720, 1440], 
		"request_mentions_intervals": [1, 3, 5, 10, 15, 30, 60], 
		"replace_regexp": /\{(\w+)\}/g, 
		"retry_count": 10, 
		"clear_test_result_interval": 5, 
		"available_keys": {
			"name": 1, "screen_name": 1, "location": 1, "description": 1, "friends_count": 1, "followers_count": 1, "statuses_count": 1, 
			"year": 1, "month": 1, "day": 1, "hour": 1, "minute": 1
		}, 
		"weekdays": {"sun": 0, "mon": 1, "tue": 2, "wed": 3, "thu": 4, "fri": 5, "sat": 6}, 
		"scheduled_tweet_snippets": [
			{"description": "Code Snippets", "content": ""}, 
			{"description": "Tweet at the specified time", "content": "\n?[12:00] It's noon!\n"}, 
			{"description": "Tweet randomly at the specified time", "content": "\n?[12:00]\nIt's noon!\nLet's have lunch!\nI'm famished!\n\n"}, 
			{"description": "Tweet every hour at the hour and at the half hour", "content": "\n?[*:00,*:30] It's time!\n"}, 
			{"description": "Tweet at the specified date and time", "content": "\n?[1/1 0:00] Happy New Year!\n"}, 
			{"description": "Tweet randomly at the specified date and time", "content": "\n?[1/1 0:00]\nHappy New Year!\nIt's the beginning of another year!\nLet's make this a good year!\n\n"}, 
			{"description": "Tweet at the first day of every month", "content": "\n?[*/1 0:00] It's the first day of the month!\n"}, 
			{}
		], 
		"reply_snippets": [
			{"description": "Code Snippets", "content": ""}, 
			{"description": "Regular expression conditions", "content": [
				{"description": "Reply to the mention includes specified words (case sensitive)", "content": "\n?/hello|hi/ Hi there!\n"}, 
				{"description": "Reply to the mention includes specified words (case insensitive)", "content": "\n?/hello|hi/i Hi there!\n"}, 
				{"description": "Reply randomly to the mention includes specified words (case sensitive)", "content": "\n?/hello|hi/\nHi again!\nHow are you?\nWhat's up?\n\n"}, 
				{"description": "Reply randomly to the mention includes specified words (case insensitive)", "content": "\n?/hello|hi/i\nHi again!\nHow are you?\nWhat's up?\n\n"}, 
				{}
			]}, 
			{"description": "User attribute conditions", "content": [
				{"description": "Reply when the user's location is specified value", "content": "\n?(location==Tokyo) You are in Tokyo!\n"}, 
				{"description": "Reply when the user's bio includes specified words", "content": "\n?(description=~/apple|mac|ipod|iphone|ipad/i) Are you an Apple fan?\n"}, 
				{"description": "Reply when the user's followers count is equal to or larger than the specified value", "content": "\n?(followers_count>=1000) You have many followers!\n"}, 
				{"description": "Reply when the user's following count is equal to or larger than the specified value", "content": "\n?(friends_count>=1000) You follow many people!\n"}, 
				{"description": "Reply when the user's tweets count is smaller than the specified value", "content": "\n?(statuses_count<1000) Hey, tweet more!\n"}, 
				{}
			]}, 
			{"description": "Date and time conditions", "content": [
				{"description": "Change reply according to the date (case sensitive)", "content": "\n?/hello|hi/\n?[1/1] Happy New Year!\n?[10/31] Trick or Treat!\n?[12/1-12/31] Happy Holidays!\nHello!\n\n"}, 
				{"description": "Change reply according to the date (case insensitive)", "content": "\n?/hello|hi/i\n?[1/1] Happy New Year!\n?[10/31] Trick or Treat!\n?[12/1-12/31] Happy Holidays!\nHello!\n\n"}, 
				{"description": "Change reply according to the time (case sensitive)", "content": "\n?/hello|hi/\n?[4-11] Good morning!\n?[12-16] Good afternoon!\n?[17-21] Good evening!\nGood night!\n\n"}, 
				{"description": "Change reply according to the time (case insensitive)", "content": "\n?/hello|hi/i\n?[4-11] Good morning!\n?[12-16] Good afternoon!\n?[17-21] Good evening!\nGood night!\n\n"}, 
				{}
			]}, 
			{"description": "Replacement", "content": [
				{"description": "{name}: Name", "content": "{name}"}, 
				{"description": "{screen_name}: Username", "content": "{screen_name}"}, 
				{"description": "{location}: Location", "content": "{location}"}, 
				{"description": "{description}: Bio", "content": "{description}"}, 
				{"description": "{followers_count}: Followers count", "content": "{followers_count}"}, 
				{"description": "{friends_count}: Following count", "content": "{friends_count}"}, 
				{"description": "{statuses_count}: Tweets count", "content": "{statuses_count}"}, 
				{}
			]}, 
			{"description": "Backreferences for regular expressions", "content": [
				{"description": "{0}: Entire matched string", "content": "{0}"}, 
				{"description": "{1}: Part of matched string corresponding to the first parentheses", "content": "{1}"}, 
				{"description": "{2}: Part of matched string corresponding to the second parentheses", "content": "{2}"}, 
				{"description": "{3}: Part of matched string corresponding to the third parentheses", "content": "{3}"}, 
				{"description": "{4}: Part of matched string corresponding to the fourth parentheses", "content": "{4}"}, 
				{"description": "{5}: Part of matched string corresponding to the fifth parentheses", "content": "{5}"}, 
				{"description": "{6}: Part of matched string corresponding to the sixth parentheses", "content": "{6}"}, 
				{"description": "{7}: Part of matched string corresponding to the seventh parentheses", "content": "{7}"}, 
				{"description": "{8}: Part of matched string corresponding to the eighth parentheses", "content": "{8}"}, 
				{"description": "{9}: Part of matched string corresponding to the nineth parentheses", "content": "{9}"}, 
				{}
			]}, 
			{"description": "Regular expressions", "content": [
				{"description": "\\s+: Spaces", "content": "\\s+"}, 
				{"description": "\\d+: Numbers", "content": "\\d+"}, 
				{"description": "[A-Za-z]+: Alphabets", "content": "[A-Za-z]+"}, 
				{"description": "[ぁ-んー]+: Hiragana", "content": "[ぁ-んー]+"}, 
				{"description": "[ァ-ヶー]+: Katakana", "content": "[ァ-ヶー]+"}, 
				{}
			]}, 
			{}
		], 
		"max_status_length": 140, 
		"max_recent_accounts_count": 20, 
		"check_for_updates_url": "http://www.mybdesign.com/widgets/db/latest_version.txt", 
		"check_for_updates_interval": 1440, 
		"support_url": "http://www.mybdesign.com/widgets/db/", 
		"err_failed_to_connect_to_twitter": "Failed to connect to Twitter.", 
		"max_logs_count": 100, 
		"system_log": "<SYSTEM>", 
		"error_log": "<ERROR>", 
		"debug_log": "<DEBUG>", 
		"debug": false, 
		"": ""
	};
	
	var prefs = {
		"active": false, 
		"expanded": false, 
		"oauth_token": "", 
		"name": "", 
		"image": "", 
		"tweet_contents": "//// 通常ツイート\n// Tips\n「//」で始まる行はコメントとみなされ、ツイートとしては無視されます。\n各ツイートの指定は、一行が一ツイートです。\n「ツイート」タブで、定期的に投稿するツイートを設定します。\nツイートの種類は、「ランダム」と「連続」があります。「連続」を選ぶと先頭から順に投稿して、最後まで到達するとまた先頭から始まります。\n「時間指定ツイート」タブで、通常のツイートとは別に指定した時間に投稿するツイートを設定します。\n時間指定ツイートでは、必ず日時条件を指定する必要があります。\n「返信」タブで、自分宛てのツイートに対しての返信を設定します。\n返信では、正規表現条件、日時条件、ユーザー属性条件が使用できます。\n時間指定ツイートや返信では、条件構文を使って条件によって投稿する内容を制御できます。\n時間指定ツイートや返信では、コードスニペットを使って条件構文の例文を挿入することができます。\n条件構文には、一行で条件文とそれに対応するツイートを指定する「一行構文」と、一つの条件文に対し複数のツイート指定する「ブロック構文」があります。\n一行構文では、「?<条件> <ツイート>」のように書きます。\nブロック構文では、「?<条件>」と条件文を書いて改行した後、続けてツイートを一行ずつ書いていきます。何も書かず改行すると、そこでブロックは終わります。\nブロック構文では、条件文に合致した場合に投稿する内容はランダムに選択されます。\nブロック構文では、条件文の次の行にさらに条件文を書くことで、条件を入れ子(ネスト)にすることができます。\n正規表現条件は、返信で使用できます。\n正規表現条件は、以下の書式です: ?/<正規表現>/\n正規表現条件で、「?/<正規表現>/i」のように「i」オプションを指定すると、大文字・小文字を区別せずにマッチします。\n正規表現条件でツイートに「{n}」を指定すると、正規表現中のn番目の括弧に対応する後方参照として置換されます。\n日時条件は、時間指定ツイート、返信で使用できます。\n日時条件は、以下の書式です: ?[YYYY/MM/DD hh:mm]\n日時条件で使用する時刻は、24時間表記です。\n日時条件では、「?[2010/1/2 3:04]」のように日時を指定します。\n日時条件では、「?[2010/1/2 3:04-2010/5/6 7:08]」のように日時の範囲を指定できます。\n日時条件では、「?[2010/1/2-2010/3/4]」のように日付の範囲を指定できます。\n日時条件では、「?[1:02-3:04]」のように時刻の範囲を指定できます。\n日時条件では、「?[1-23]」のように時刻の範囲を指定できます。\n日時条件では、カンマ区切りで複数の範囲をしてできます。\n日時条件では、ワイルドカードとして「*」を使用できます。\nユーザー属性条件は、返信で使用できます。\nユーザー属性条件は、以下の書式です: ?(<属性キー><演算子><値>)\nユーザー属性条件で使用できる属性キーは、name、screen_name、location、description、friends_count、followers_count、statuses_countです。\nユーザー属性条件で使用できる演算子は、==、!=、<、<=、>、>=です。また、「?(<属性キー>=~/.../)」のような形式で正規表現にマッチさせることができます。\n「ブロックしたユーザー」タブでは、返信しないユーザーを設定します。\n自分自身からのツイートには返信しません。\n", 
		"scheduled_tweet_contents": "//// 時間指定ツイート\n// 時報\n?[1/1 0:00]\n謹賀新年！\nあけましておめでとうございます。\n今年もよろしくお願いします。\n\n\n?[0:00,3:00,6:00,9:00,12:00,15:00,18:00,21:00]\n{hour}時です。\n{hour}時ですよ。\n{hour}時になりました。\n\n", 
		"reply_contents": "//// 返信\n// Tips\n?/正規表現.*(教|おし)えて/\n正規表現についてはこのサイトが参考になりますよ。 http://www.mnet.ne.jp/~nakama/\n正規表現条件は、以下の書式です: ?/<正規表現>/\n\n\n?/構文.*(教|おし)えて/\n条件構文には、一行で条件文とそれに対応するツイートを指定する「一行構文」と、一つの条件文に対し複数のツイート指定する「ブロック構文」があります。\n一行構文では、「?<条件> <ツイート>」のように書きます。\nブロック構文では、「?<条件>」と条件文を書いて改行した後、続けてツイートを一行ずつ書いていきます。何も書かず改行すると、そこでブロックは終わります。\nブロック構文では、条件文に合致した場合に投稿する内容はランダムに選択されます。\nブロック構文では、条件文の次の行にさらに条件文を書くことで、条件を入れ子(ネスト)にすることができます。\n正規表現条件は、返信で使用できます。\n正規表現条件は、以下の書式です: ?/<正規表現>/\n正規表現条件で、「?/<正規表現>/i」のように「i」オプションを指定すると、大文字・小文字を区別せずにマッチします。\n正規表現条件でツイートに「{n}」を指定すると、正規表現中のn番目の括弧に対応する後方参照として置換されます。\n日時条件は、時間指定ツイート、返信で使用できます。\n日時条件は、以下の書式です: ?[YYYY/MM/DD hh:mm]\n日時条件で使用する時刻は、24時間表記です。\n日時条件では、「?[2010/1/2 3:04]」のように日時を指定します。\n日時条件では、「?[2010/1/2 3:04-2010/5/6 7:08]」のように日時の範囲を指定できます。\n日時条件では、「?[2010/1/2-2010/3/4]」のように日付の範囲を指定できます。\n日時条件では、「?[1:02-3:04]」のように時刻の範囲を指定できます。\n日時条件では、「?[1-23]」のように時刻の範囲を指定できます。\n日時条件では、カンマ区切りで複数の範囲をしてできます。\n日時条件では、ワイルドカードとして「*」を使用できます。\nユーザー属性条件は、返信で使用できます。\nユーザー属性条件は、以下の書式です: ?(<属性キー><演算子><値>)\nユーザー属性条件で使用できる属性キーは、name、screen_name、location、description、friends_count、followers_count、statuses_countです。\nユーザー属性条件で使用できる演算子は、==、!=、<、<=、>、>=です。また、「?(<属性キー>=~/.../)」のような形式で正規表現にマッチさせることができます。\n\n\n// 挨拶\n?/おはよ(う|ー)|オハヨ(ウ|ー)/\n?[3-6] お早いですね！\n?[7-10] おはようございます。\n?[11-14] もうお昼ですよ。\n?[15-17] もう夕方ですよ。\nこんな時間にお目覚めですか？\n\n\n?/こんにち(は|わ)|コンニチ(ハ|ワ)/\n?[18-23,0-3] こんばんは。\n?[4-10] おはようございます。\nこんにちは！\n\n\n?/こんばん(は|わ)|コンバン(ハ|ワ)/\n?[18-23,0-3] こんばんは！\nえっ？\n\n\n// 時刻\n?/何時(？|\?|か(教|おし)えて)/\n?[*] {hour}時{minute}分ぐらいです。\nうーん、分かりません。\n\n\n// 以下ランダムな返答\nうーん、よく分かりません。\n{name}さんは{location}にいらっしゃるのですか？\n{name}さんは{friends_count}人フォローしていますね。\n{name}さんはフォロワーが{followers_count}人いますね。\n", 
		"ignore_words": "RT @, QT @", 
		"blocked_users": "", 
		"tweet_type": 1, 
		"tweet_sequence": 0, 
		"tweet_interval": 60, 
		"request_mentions_interval": 3, 
		"last_tweeted": "",
		"last_tweeted_time": 0, 
		"last_scheduled_tweeted": "",
		"last_scheduled_tweeted_time": 0, 
		"last_checked_for_updates_time": 0, 
		"last_id": "", 
		"sleep_until": 0, 
		"selected_tab": "tweets_tab", 
		"": ""
	};
	
	var recentAccounts = [];
	
	var prefKeys = {};
	
	var session = {
		"launched_date": null, 
		"flipped": false, 
		"error_shown": false, 
		"controls_shown": false, 
		"active_status": false, 
		"main_loop_observer": null, 
		"wait_for_request_mentions": 0, 
		"user": {}, 
		"tweeting": false, 
		"replying": false, 
		"tweet_contents": {}, 
		"scheduled_tweet_contents": {}, 
		"reply_contents": {}, 
		"ignore_words": "", 
		"blocked_users": {}, 
		"mentions": [], 
		"mentions_dict": {}, 
		"test_tweet_sequence": 0,  
		"test_last_tweeted": "", 
		"test_last_scheduled_tweeted": "", 
		"last_verified_time": 0, 
		"migrating_from": null, 
		"front_out_delay": null, 
		"clear_test_result_delay": null, 
		"show_controls_animator": null, 
		"hide_controls_animator": null, 
		"expand_animator": null, 
		"collapse_animator": null, 
		"show_log_animator": null, 
		"hide_log_animator": null, 
		"turn_on_indicator_animator": null, 
		"turn_off_indicator_animator": null, 
		"show_error_animator": null, 
		"hide_error_animator": null, 
		"": ""
	};
	
	var elements = {};
	
	
	//////// Methods
	self.initialize = function()
	{
		self.setUpControls();
		self.setUpPreferenceKeys();
		
		if(window.widget)
		{
			widget.onshow = self.willShow;
			widget.onhide = self.willHide;
			
			// Load recent accounts
			self.loadRecentAccounts();
			if(recentAccounts.length > 0)
			{
				self.setUpRecentAccountsSelections();
				document.getElementById("choose_from_recent_accounts").style.display = "block";
			}
		}
		
		self.launch(true);
	};
	
	
	self.launch = function(startIfActive)
	{
		if(session.migrating_from)
		{
			self.setUpPreferenceKeys(session.migrating_from);
		}
		
		self.loadPreferences();
		self.restorePreferences();
		
		if(session.migrating_from)
		{
			self.setUpPreferenceKeys();
			self.savePreferences();
		}
		
		if(window.dashbot_plugin)
		{
			if(prefs.oauth_token != "" && dashbot_plugin.isAuthorized(prefs.oauth_token))
			{
				elements.unauthorized.style.display = "none";
				elements.authorizing.style.display = "none";
				elements.main.style.display = "block";
				if(self.setUpTweets(prefs.tweet_contents, prefs.scheduled_tweet_contents, prefs.reply_contents))
				{
					if(startIfActive && prefs.active)
					{
						window.setTimeout(function(){ self.startMainLoop(); }, 2000);
					}
				}
				self.setUpIgnoreWords();
				self.setUpBlockedUsers();
			}
			else
			{
				// Not authorized
				elements.unauthorized.style.display = "block";
				elements.authorizing.style.display = "none";
				elements.main.style.display = "none";
			}
		}
		
		if(prefs.expanded)
		{
			window.setTimeout(function(){ prefs.expanded = false; self.expand(); }, 0);
		}
		
		self.displayTweetSequence(prefs.tweet_sequence, session.tweet_contents);
		self.displayTestTweetSequence(session.test_tweet_sequence, elements.tweet_type2.checked? session.tweet_contents: null);
		
		session.launched_date = new Date;
	};
	
	
	self.setUpControls = function()
	{
		//// Front
		// Set up scroll area
		elements.log = document.getElementById("log");
		elements.log_content = document.getElementById("log_content");
		elements.log_scrollbar = new AppleVerticalScrollbar(document.getElementById("log_scrollbar"));
		elements.log_scrollbar.minThumbSize = 19;
		elements.log_scroll_area = new AppleScrollArea(elements.log_content);
		elements.log_scroll_area.addScrollbar(elements.log_scrollbar);
		
		// Set up buttons
		document.getElementById("connect_button").addEventListener("click", self.didClickConnectButton);
		document.getElementById("enter_button").addEventListener("click", self.didClickEnterButton);
		document.getElementById("cancel_enter_button").addEventListener("click", self.didClickCancelEnterButton);
		elements.start_button = document.getElementById("start_button");
		elements.start_button.addEventListener("click", self.didClickStartButton);
		elements.stop_button = document.getElementById("stop_button");
		elements.stop_button.addEventListener("click", self.didClickStopButton);
		elements.indicator = document.getElementById("indicator");
		elements.expand_button = document.getElementById("expand_button");
		elements.expand_button.addEventListener("click", self.didClickExpandButton);
		elements.collapse_button = document.getElementById("collapse_button");
		elements.collapse_button.addEventListener("click", self.didClickCollapseButton);
		elements.info_button = new AppleInfoButton(
			document.getElementById("info_button"), document.getElementById("front"), 
			"black", "black", self.didClickInfoButton);
		
		// Other controls
		elements.front = document.getElementById("front");
		elements.front.addEventListener("mouseover", function(e){ self.showControls(); }, false);
		elements.front.addEventListener("mouseout", function(e){ self.hideControls(); }, false);
		elements.front_right = document.getElementById("front_right");
		elements.unauthorized = document.getElementById("unauthorized");
		elements.recent_accounts = document.getElementById("recent_accounts");
		elements.recent_accounts.addEventListener("change", self.didChangeRecentAccount, false);
		elements.authorizing = document.getElementById("authorizing");
		elements.main = document.getElementById("main");
		elements.name = document.getElementById("name");
		elements.api_remaining = document.getElementById("api_remaining");
		elements.loading = document.getElementById("loading");
		elements.image = document.getElementById("image");
		
		//// Back
		// Set up tabs
		elements.tabs = document.getElementById("tabs").getElementsByTagName("li");
		for(var i=0; i<elements.tabs.length; i++)
		{
			elements.tabs[i].addEventListener("mouseover", self.didMouseOverTab);
			elements.tabs[i].addEventListener("mouseout", self.didMouseOutTab);
			elements.tabs[i].addEventListener("click", self.didClickTab);
		}
		
		// Set up buttons
		elements.tweet_test_button = document.getElementById("tweet_test_button");
		elements.tweet_test_button.value = L("Test");
		elements.tweet_test_button.addEventListener("click", self.didClickTweetTestButton);
		elements.reset_sequence_button = document.getElementById("reset_sequence_button");
		elements.reset_sequence_button.value = L("Reset");
		elements.reset_sequence_button.addEventListener("click", self.didClickResetSequenceButton);
		document.getElementById("scheduled_tweet_add_snippet_button").addEventListener("click", self.didClickScheduledTweetAddSnippetButton);
		elements.scheduled_tweet_test_button = document.getElementById("scheduled_tweet_test_button");
		elements.scheduled_tweet_test_button.value = L("Test");
		elements.scheduled_tweet_test_button.addEventListener("click", self.didClickScheduledTweetTestButton);
		document.getElementById("reply_add_snippet_button").addEventListener("click", self.didClickReplyAddSnippetButton);
		elements.reply_test_button = document.getElementById("reply_test_button");
		elements.reply_test_button.value = L("Test");
		elements.reply_test_button.addEventListener("click", self.didClickReplyTestButton);
		new AppleGlassButton(document.getElementById("done_button"), L("Done"), self.didClickDoneButton);
		new AppleGlassButton(document.getElementById("cancel_button"), L("Cancel"), self.didClickCancelButton);
		new AppleGlassButton(document.getElementById("clear_log_button"), L("Clear Log"), self.didClickClearLogButton);
		
		// Copyright
		document.getElementById("version").innerHTML = config.version;
		
		// Other controls
		elements.back = document.getElementById("back");
		
		elements.tweet_contents = document.getElementById("tweet_contents");
		elements.test_tweet_sequence = document.getElementById("test_tweet_sequence");
		elements.tweet_test_result = document.getElementById("tweet_test_result");
		elements.tweet_type1 = document.getElementById("tweet_type1");
		elements.tweet_type1.addEventListener("click", self.didChangeTweetType);
		elements.tweet_type2 = document.getElementById("tweet_type2");
		elements.tweet_type2.addEventListener("click", self.didChangeTweetType);
		elements.tweet_sequence = document.getElementById("tweet_sequence");
		elements.tweet_interval = document.getElementById("tweet_interval");
		self.setUpSelections(elements.tweet_interval, config.tweet_intervals, self.stringFromMinutes, 1);
		
		elements.scheduled_tweet_contents = document.getElementById("scheduled_tweet_contents");
		elements.scheduled_tweet_snippets = document.getElementById("scheduled_tweet_snippets");
		self.setUpSelections(elements.scheduled_tweet_snippets, config.scheduled_tweet_snippets);
		elements.scheduled_tweet_test_year = document.getElementById("scheduled_tweet_test_year");
		elements.scheduled_tweet_test_month = document.getElementById("scheduled_tweet_test_month");
		elements.scheduled_tweet_test_day = document.getElementById("scheduled_tweet_test_day");
		elements.scheduled_tweet_test_hour = document.getElementById("scheduled_tweet_test_hour");
		elements.scheduled_tweet_test_minute = document.getElementById("scheduled_tweet_test_minute");
		self.setUpDateConditions();
		elements.scheduled_tweet_test = document.getElementById("scheduled_tweet_test");
		elements.scheduled_tweet_test_result = document.getElementById("scheduled_tweet_test_result");
		
		elements.reply_contents = document.getElementById("reply_contents");
		elements.reply_snippets = document.getElementById("reply_snippets");
		self.setUpSelections(elements.reply_snippets, config.reply_snippets);
		elements.reply_test = document.getElementById("reply_test");
		elements.reply_test_result = document.getElementById("reply_test_result");
		elements.request_mentions_interval = document.getElementById("request_mentions_interval");
		self.setUpSelections(elements.request_mentions_interval, config.request_mentions_intervals, self.stringFromMinutes);
		
		elements.ignore_words = document.getElementById("ignore_words");
		elements.blocked_users = document.getElementById("blocked_users");
	};
	
	
	self.setUpDateConditions = function()
	{
		var now = self.deconstructDate(new Date);
		
		var years = new Array(10);
		for(var i=0; i<years.length; i++)
			years[i] = now.year + i;
		self.setUpSelections(elements.scheduled_tweet_test_year, years);
		elements.scheduled_tweet_test_year.value = now.year;
		
		var months = new Array(12);
		for(var i=0; i<months.length; i++)
			months[i] = i + 1;
		self.setUpSelections(elements.scheduled_tweet_test_month, months);
		elements.scheduled_tweet_test_month.value = now.month;
		
		var days = new Array(31);
		for(var i=0; i<days.length; i++)
			days[i] = i + 1;
		self.setUpSelections(elements.scheduled_tweet_test_day, days);
		elements.scheduled_tweet_test_day.value = now.day;
		
		var hours = new Array(24);
		for(var i=0; i<hours.length; i++)
			hours[i] = i;
		self.setUpSelections(elements.scheduled_tweet_test_hour, hours, function(text){ return self.zeroFill(text, 2); });
		elements.scheduled_tweet_test_hour.selectedIndex = now.hour;
		
		var minutes = new Array(60);
		for(var i=0; i<minutes.length; i++)
			minutes[i] = i;
		self.setUpSelections(elements.scheduled_tweet_test_minute, minutes, function(text){ return self.zeroFill(text, 2); });
		elements.scheduled_tweet_test_minute.selectedIndex = 0;
	};
	
	
	self.setUpSelections = function(target, selections, callback, debugSkipCount)
	{
		target.innerHTML = "";
		
		var option, description, content;
		for(var i=0; i<selections.length; i++)
		{
			if(!config.debug && i < debugSkipCount)
				continue;
			
			if(typeof(selections[i]) == "object")
			{
				description = L(selections[i].description);
				content = L(selections[i].content);
			}
			else
			{
				description = content = L(selections[i]);
			}
			
			if(self.isUndefined(description) || self.isUndefined(content))
				continue;
			
			if(typeof(content) == "object")
			{
				// Group
				option = document.createElement("optgroup");
				option.label = description;
				self.setUpSelections(option, content, callback);
			}
			else
			{
				// Option
				option = document.createElement("option");
				option.value = content;
				option.appendChild(document.createTextNode(callback? callback(description): description));
			}
			target.appendChild(option);
		}
	};
	
	
	self.setUpPreferenceKeys = function(identifier)
	{
		for(var key in prefs)
		{
			if(!key)
				continue;
			
			if(identifier != null)
				prefKeys[key] = ((identifier != "")? identifier + "-": "") + key;
//			else if(window.dashcode && !config.debug)
			else if(window.dashcode)
				prefKeys[key] = dashcode.createInstancePreferenceKey(key);
			else
				prefKeys[key] = key;
		}
	};
	
	
	self.loadPreferences = function()
	{
		if(!window.widget)
			return;
		
		var value;
		for(var key in prefs)
		{
			if(key == "")
			{
				continue;
			}
			
			if(!self.isUndefined(value = widget.preferenceForKey(prefKeys[key])))
			{
				prefs[key] = value;
			}
		}
	};
	
	
	self.restorePreferences = function()
	{
		elements.tweet_contents.value = prefs.tweet_contents;
		elements.tweet_type1.checked = (prefs.tweet_type == 1);
		elements.tweet_type2.checked = (prefs.tweet_type == 2);
		elements.reset_sequence_button.disabled = !elements.tweet_type2.checked;
		elements.tweet_interval.value = prefs.tweet_interval;
		elements.scheduled_tweet_contents.value = prefs.scheduled_tweet_contents;
		elements.reply_contents.value = prefs.reply_contents;
		elements.request_mentions_interval.value = prefs.request_mentions_interval;
		elements.ignore_words.value = prefs.ignore_words;
		elements.blocked_users.value = prefs.blocked_users;
	};
	
	
	self.savePreferences = function()
	{
		if(!window.widget)
			return;
		
		for(var key in prefs)
		{
			if(!key || !prefKeys[key])
				continue;
			
			widget.setPreferenceForKey(prefs[key], prefKeys[key]);
		}
	};
	
	
	self.storePreferences = function()
	{
		if(!self.setUpTweets(elements.tweet_contents.value, elements.scheduled_tweet_contents.value, elements.reply_contents.value))
			return false;
		
		self.setUpIgnoreWords();
		self.setUpBlockedUsers();
		
		if(prefs.tweet_contents != elements.tweet_contents.value)
			self.resetSequence();
		
		prefs.tweet_contents = elements.tweet_contents.value;
		prefs.tweet_type = (elements.tweet_type1.checked? 1: (elements.tweet_type2.checked? 2: 0));
		prefs.tweet_interval = elements.tweet_interval.value;
		prefs.scheduled_tweet_contents = elements.scheduled_tweet_contents.value;
		prefs.reply_contents = elements.reply_contents.value;
		prefs.request_mentions_interval = elements.request_mentions_interval.value;
		prefs.ignore_words = elements.ignore_words.value;
		prefs.blocked_users = elements.blocked_users.value;
		
		return true;
	};
	
	
	self.removePreferencesFor = function(identifier)
	{
		if(!window.widget)
			return;
		
		self.setUpPreferenceKeys(identifier);
		for(var key in prefs)
		{
			if(!key || !prefKeys[key])
				continue;
			
			widget.setPreferenceForKey(null, prefKeys[key]);
		}
		self.setUpPreferenceKeys();
	};
	
	
	self.loadRecentAccounts = function()
	{
		if(!window.widget)
			return;
		
		var identifier, name;
		for(var i=0; i<config.max_recent_accounts_count; i++)
		{
			identifier = widget.preferenceForKey("recent_account_" + (i + 1) + "_identifier");
			name = widget.preferenceForKey("recent_account_" + (i + 1) + "_name");
			if(identifier != null && name != null)
				recentAccounts.push({"identifier": identifier, "name": name});
		}
	};


	self.saveRecentAccounts = function()
	{
		if(!window.widget)
			return;
		
		var identifier, name;
		for(var i=0; i<config.max_recent_accounts_count; i++)
		{
			identifier = ((i < recentAccounts.length)? recentAccounts[i].identifier: null);
			name = ((i < recentAccounts.length)? recentAccounts[i].name: null);
			widget.setPreferenceForKey(identifier, "recent_account_" + (i + 1) + "_identifier");
			widget.setPreferenceForKey(name, "recent_account_" + (i + 1) + "_name");
		}
	};
	
	
	//// Animations
	self.showPreferences = function()
	{
		window.resizeBy((prefs.expanded? 0: config.expanded_size.width - config.collapsed_size.width), 
			config.back_size.height - (prefs.expanded? config.expanded_size.height: config.collapsed_size.height));
		
		if(window.widget)
			widget.prepareForTransition("ToBack");
		
		session.flipped = true;
		
		elements.front.style.display = "none";
		elements.back.style.display = "block";
		
		if(prefs.selected_tab != "")
			self.selectTab(document.getElementById(prefs.selected_tab));
		
		if(window.widget)
			window.setTimeout(function(){ widget.performTransition(); }, 0);
	};


	self.hidePreferences = function()
	{
		if(window.widget)
			widget.prepareForTransition("ToFront");
		
		session.flipped = false;
		
		elements.front.style.display = "block";
		elements.back.style.display = "none";
		
		if(window.widget)
			window.setTimeout(function(){ widget.performTransition(); }, 0);
		
		window.resizeBy((prefs.expanded? 0: config.collapsed_size.width - config.expanded_size.width), 
			(prefs.expanded? config.expanded_size.height: config.collapsed_size.height) - config.back_size.height);
		
		elements.info_button._frontOutDelay();
		elements.info_button._labelOut();
	};
	
	
	self.showControls = function()
	{
		if(session.front_out_delay)
		{
			window.clearTimeout(session.front_out_delay);
			delete session.front_out_delay;
			session.front_out_delay = null;
		}
		
		if(session.controls_shown)
			return;
		
		if(session.hide_controls_animator)
		{
			session.hide_controls_animator.stop();
			delete session.hide_controls_animator;
			session.hide_controls_animator = null;
		}
		
		session.controls_shown = true;
		
		if(!prefs.expanded)
		{
			elements.expand_button.style.opacity = 0;
			elements.expand_button.style.display = "block";
		}
		else
		{
			elements.collapse_button.style.opacity = 0;
			elements.collapse_button.style.display = "block";
		}
		
		elements.api_remaining.style.opacity = 0;
		elements.api_remaining.style.display = "block";
		
		session.show_controls_animator = new AppleAnimator(500, 13);
		session.show_controls_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(0, 1, self.animationHandler);
		session.show_controls_animator.addAnimation(animation);
		session.show_controls_animator.start();
	};
	
	
	self.hideControls = function()
	{
		if(!session.front_out_delay)
		{
			session.front_out_delay = window.setTimeout(function(){ self.hideControls(); }, 0);
			return;
		}
		
		if(!session.controls_shown)
			return;
		
		if(session.show_controls_animator)
		{
			session.show_controls_animator.stop();
			delete session.show_controls_animator;
			session.show_controls_animator = null;
		}
		
		session.controls_shown = false;
		
		elements.name.style.opacity = 0;
		elements.name.style.display = "block";
		
		session.hide_controls_animator = new AppleAnimator(500, 13);
		session.hide_controls_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(1, 0, self.animationHandler);
		session.hide_controls_animator.addAnimation(animation);
		session.hide_controls_animator.start();
	};
	
	
	self.expand = function()
	{
		if(prefs.expanded)
			return;
		
		if(session.collapse_animator)
		{
			session.collapse_animator.stop();
			delete session.collapse_animator;
			session.collapse_animator = null;
		}
		
		prefs.expanded = true;
		if(window.widget)
			widget.setPreferenceForKey(prefs.expanded, prefKeys.expanded);
		
		window.resizeBy(config.expanded_size.width - config.collapsed_size.width, 0);
		elements.expand_button.style.display = "none";
		elements.collapse_button.style.display = "block";
		
		session.expand_animator = new AppleAnimator(300, 13);
		session.expand_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(config.collapsed_size.width, config.expanded_size.width, self.animationHandler);
		session.expand_animator.addAnimation(animation);
		session.expand_animator.start();
	};
	
	
	self.collapse = function()
	{
		if(!prefs.expanded)
			return;
		
		if(session.expand_animator)
		{
			session.expand_animator.stop();
			delete session.expand_animator;
			session.expand_animator = null;
		}
		
		prefs.expanded = false;
		if(window.widget)
			widget.setPreferenceForKey(prefs.expanded, prefKeys.expanded);
		
		session.collapse_animator = new AppleAnimator(300, 13);
		session.collapse_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(config.expanded_size.width, config.collapsed_size.width, self.animationHandler);
		session.collapse_animator.addAnimation(animation);
		session.collapse_animator.start();
	};
	
	
	self.showLog = function()
	{
		elements.log.style.opacity = 0;
		elements.log.style.display = "block";
		
		session.show_log_animator = new AppleAnimator(200, 13);
		session.show_log_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(0, 1, self.animationHandler);
		session.show_log_animator.addAnimation(animation);
		session.show_log_animator.start();
		
		self.refreshLog();
	};
	
	
	self.hideLog = function()
	{
		session.hide_log_animator = new AppleAnimator(200, 13);
		session.hide_log_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(1, 0, self.animationHandler);
		session.hide_log_animator.addAnimation(animation);
		session.hide_log_animator.start();
	};
	
	
	self.turnOnIndicator = function()
	{
		if(session.turn_off_indicator_animator)
		{
			session.turn_off_indicator_animator.stop();
			delete session.turn_off_indicator_animator;
			session.turn_off_indicator_animator = null;
		}
		
		session.turn_on_indicator_animator = new AppleAnimator(2000, 13);
		session.turn_on_indicator_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(0, 1, self.animationHandler);
		session.turn_on_indicator_animator.addAnimation(animation);
		session.turn_on_indicator_animator.start();
	};
	
	
	self.turnOffIndicator = function()
	{
		if(session.turn_on_indicator_animator)
		{
			session.turn_on_indicator_animator.stop();
			delete session.turn_on_indicator_animator;
			session.turn_on_indicator_animator = null;
		}
		
		session.turn_off_indicator_animator = new AppleAnimator(2000, 13);
		session.turn_off_indicator_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(1, 0, self.animationHandler);
		session.turn_off_indicator_animator.addAnimation(animation);
		session.turn_off_indicator_animator.start();
	};
	
	
	self.showError = function()
	{
		if(session.error_shown)
			return;
		
		if(!elements.error)
			return;
		
		session.error_shown = true;
		
		elements.error.style.opacity = 0;
		elements.error.style.display = "block";
		
		session.show_error_animator = new AppleAnimator(200, 13);
		session.show_error_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(0, 1, self.animationHandler);
		session.show_error_animator.addAnimation(animation);
		session.show_error_animator.start();
	};
	
	
	self.hideError = function()
	{
		if(!session.error_shown)
			return;
		
		session.error_shown = false;
		
		session.hide_error_animator = new AppleAnimator(200, 13);
		session.hide_error_animator.oncomplete = self.animationComplete;
		var animation = new AppleAnimation(1, 0, self.animationHandler);
		session.hide_error_animator.addAnimation(animation);
		session.hide_error_animator.start();
	};
	
	
	self.animationHandler = function(animator, current, start, finish)
	{
		var animation = this;
		if(animator == session.show_controls_animator)
		{
			if(!prefs.expanded)
			{
				elements.expand_button.style.opacity = current;
			}
			else
			{
				elements.collapse_button.style.opacity = current;
			}
			
			elements.name.style.opacity = 1 - current;
			elements.api_remaining.style.opacity = current;
		}
		else if(animator == session.hide_controls_animator)
		{
			if(!prefs.expanded)
			{
				elements.expand_button.style.opacity = current;
			}
			else
			{
				elements.collapse_button.style.opacity = current;
			}
			
			elements.name.style.opacity = 1 - current;
			elements.api_remaining.style.opacity = current;
		}
		else if(animator == session.expand_animator)
		{
			elements.front_right.style.width = Math.floor(current - (config.collapsed_size.width - config.side_margin)) + "px";
		}
		else if(animator == session.collapse_animator)
		{
			elements.front_right.style.width = Math.floor(current - (config.collapsed_size.width - config.side_margin)) + "px";
		}
		else if(animator == session.show_log_animator)
		{
			elements.log.style.opacity = current;
		}
		else if(animator == session.hide_log_animator)
		{
			elements.log.style.opacity = current;
		}
		else if(animator == session.turn_on_indicator_animator)
		{
			elements.indicator.style.opacity = current;
		}
		else if(animator == session.turn_off_indicator_animator)
		{
			elements.indicator.style.opacity = current;
		}
		else if(animator == session.show_error_animator)
		{
			elements.error.style.opacity = current;
		}
		else if(animator == session.hide_error_animator)
		{
			elements.error.style.opacity = current;
		}
	};
	
	
	self.animationComplete = function()
	{
		var animator = this;
		if(animator == session.show_controls_animator)
		{
			elements.name.style.display = "none";
			
			session.show_controls_animator = null;
		}
		else if(animator == session.hide_controls_animator)
		{
			if(!prefs.expanded)
			{
				elements.expand_button.style.display = "none";
			}
			else
			{
				elements.collapse_button.style.display = "none";
			}
			
			elements.api_remaining.style.display = "none";
			
			session.hide_controls_animator = null;
		}
		else if(animator == session.expand_animator)
		{
			session.expand_animator = null;
			self.showLog();
		}
		else if(animator == session.collapse_animator)
		{
			window.resizeBy(config.collapsed_size.width - config.expanded_size.width, 0);
			elements.expand_button.style.display = "block";
			elements.collapse_button.style.display = "none";
			
			session.collapse_animator = null;
		}
		else if(animator == session.show_log_animator)
		{
			session.show_log_animator = null;
		}
		else if(animator == session.hide_log_animator)
		{
			elements.log.style.display = "none";
			session.hide_log_animator = null;
			
			self.collapse();
		}
		else if(animator == session.turn_on_indicator_animator)
		{
			self.turnOffIndicator();
		}
		else if(animator == session.turn_off_indicator_animator)
		{
			self.turnOnIndicator();
		}
		else if(animator == session.show_error_animator)
		{
			session.show_error_animator = null;
		}
		else if(animator == session.hide_error_animator)
		{
			elements.error.parentNode.removeChild(elements.error);
			elements.error = null;
			session.hide_error_animator = null;
		}
		
		animator.stop();
		delete animator;
	};
	
	
	//// Log and errors
	self.log = function(caption, text, logClass, date)
	{
		if(!date)
		{
			date = new Date;
		}
		
		var dateComponents = self.deconstructDate(date);
		var dateString = self.zeroFill(dateComponents.year, 4) + "/" + self.zeroFill(dateComponents.month, 2) + "/" + self.zeroFill(dateComponents.day, 2);
		dateString += " " + self.zeroFill(dateComponents.hour, 2) + ":" + self.zeroFill(dateComponents.minute, 2) + ":" + self.zeroFill(dateComponents.second, 2);
		
		var item = document.createElement("div");
		item.className = "item" + (logClass? " " + logClass: "");
		
		var obj;
		obj = document.createElement("div");
		obj.className = "caption";
		obj.appendChild(document.createTextNode(caption));
		item.appendChild(obj);
		
		obj = document.createElement("div");
		obj.className = "date";
		obj.appendChild(document.createTextNode(dateString));
		item.appendChild(obj);
		
		obj = document.createElement("p");
		obj.appendChild(document.createTextNode(text));
		item.appendChild(obj);
		
		item.innerHTML = item.innerHTML.replace(/@(\w+)/g, "<a onclick=\"window.controller.openProfile('$1');\">@$1</a>");
		
		if(elements.log_content.hasChildNodes())
			elements.log_content.insertBefore(item, elements.log_content.firstChild);
		else
			elements.log_content.appendChild(item);
		
		if(elements.log_content.childNodes.length > config.max_logs_count)
		{
			for(var i=config.max_logs_count; i<elements.log_content.childNodes.length; i++)
			{
				elements.log_content.removeChild(elements.log_content.childNodes[i]);
			}
		}
		
		self.refreshLog(item);
	};
	
	
	self.clearLog = function()
	{
		elements.log_content.innerHTML = "";
		self.refreshLog();
	};
	
	
	self.refreshLog = function(item)
	{
		elements.log_scroll_area.refresh();
		if(elements.log_scrollbar.hidden)
			elements.log_content.style.width = (config.log_scroll_area_width + config.log_scrollbar_width - 2) + "px";
		else
			elements.log_content.style.width = config.log_scroll_area_width + "px";
		
		if(item)
			elements.log_scroll_area.reveal(item);
	};
	
	
	self.displayError = function(error, log, logOnly)
	{
		if(log)
		{
			self.log(config.error_log, error, "error");
		}
		
		if(session.error_shown || logOnly)
			return;
		
		var container = document.getElementById(!session.flipped? "front": "back");
		elements.error = document.createElement("div");
		elements.error.id = "error"
		container.appendChild(elements.error);
		
		var content = document.createElement("div");
		content.className = "content";
		content.innerHTML = L(error);
		elements.error.appendChild(content);
		
		var button = document.createElement("div");
		button.className = "ok_button";
		elements.error.appendChild(button);
		
		new AppleGlassButton(button, "OK", self.didClickOkButton);

		self.showError();
	};
	
	
	//// Recent accounts
	self.setUpRecentAccountsSelections = function()
	{
		var accounts = [{"description": "Recent Accounts", "content": ""}];
		for(var i=0; i<recentAccounts.length; i++)
		{
			accounts.push({"description": "@" + recentAccounts[i].name, "content": recentAccounts[i].identifier});
		}
		
		self.setUpSelections(elements.recent_accounts, accounts);
	};
	
	
	self.updateRecentAccounts = function(name, prevIdentifier)
	{
		if(!window.widget)
			return;
		
		if(!prevIdentifier)
			prevIdentifier = widget.identifier;
		
		for(var i=0; i<recentAccounts.length; i++)
		{
			if(recentAccounts[i].identifier == prevIdentifier)
			{
				recentAccounts.splice(i, 1);
				break;
			}
		}
		
		if(name != "")
			recentAccounts.unshift({"identifier": widget.identifier, "name": name});
		
		self.saveRecentAccounts();
		self.setUpRecentAccountsSelections();
	};
	
	
	self.removeFromRecentAccounts = function(identifier)
	{
		for(var i=0; i<recentAccounts.length; i++)
		{
			if(recentAccounts[i].identifier == identifier)
			{
				recentAccounts.splice(i, 1);
				break;
			}
		}
	};
	
	
	//// Display
	self.showLoadingIndicator = function()
	{
		elements.loading.style.display = "block";
	};
	
	
	self.hideLoadingIndicator= function()
	{
		elements.loading.style.display = "none";
	};
	
	
	self.setProfile = function(name, image)
	{
		prefs.name = name;
		prefs.image = image;
		if(window.widget)
		{
			widget.setPreferenceForKey(prefs.name, prefKeys.name);
			widget.setPreferenceForKey(prefs.image, prefKeys.image);
		}
		
		elements.name.innerHTML = "@" + name;
		elements.image.innerHTML = "";
		if(elements.reply_test.value == "")
			elements.reply_test.value = "@" + name + " ";
		
		if(image != "")
		{
			var img = document.createElement("img");
			img.src = image;
			img.addEventListener("click", self.didClickImage);
			elements.image.appendChild(img);
		}
	};
	
	
	self.clearProfile = function()
	{
		elements.name.innerHTML = "";
		elements.image.innerHTML = "";
		elements.reply_test.value = "";
		
		prefs.name = "";
		prefs.image = "";
		if(window.widget)
		{
			widget.setPreferenceForKey(prefs.name, prefKeys.name);
			widget.setPreferenceForKey(prefs.image, prefKeys.image);
		}
	};
	
	
	self.setApiRemaining = function(remaining_hits, hourly_limit)
	{
		elements.api_remaining.innerHTML = "API: " + remaining_hits + "/" + hourly_limit;
	};
	
	
	self.displayTweetSequence = function(sequence, contents)
	{
		var max = ((contents && contents.children)? contents.children.length: 0);
		sequence = ((max > 0)? sequence + 1: max);
		
		elements.tweet_sequence.innerHTML = (sequence? sequence: "-") + "/" + (max? max: "-");
	};
	
	
	self.displayTestTweetSequence = function(sequence, contents)
	{
		var max = ((contents && contents.children)? contents.children.length: 0);
		sequence = ((max > 0)? sequence + 1: max);
		
		elements.test_tweet_sequence.innerHTML = (sequence? sequence: "-") + "/" + (max? max: "-");
	};
	
	
	//// Parsing and checking
	self.parseContents = function(target, source, skipTweets)
	{
		var parser = new DashbotParser(self);
		var result = parser.parse(source, skipTweets);
		if(!result)
		{
			// Parse error
			self.displayError(L(target) + ": " + L("Syntax error near") + " \"" + parser.lastReadLine + "\" " + L("in line") + " " + parser.readingIndex);
			
		}
		return result;
	};
	
	
	self.setUpTweets = function(tweets, scheduledTweets, replies)
	{
		var result;
		
		// Parse tweets
		if(!(result = self.parseContents("Tweets", tweets)))
			return false;
		
		session.tweet_contents = result;
		
		// Parse scheduled tweets
		if(!(result = self.parseContents("Scheduled Tweets", scheduledTweets, true)))
			return false;
		
		session.scheduled_tweet_contents = result;
		
		// Parse replies
		if(!(result = self.parseContents("Replies", replies)))
			return false;
		
		session.reply_contents = result;
		
		return true;
	};
	
	
	self.setUpIgnoreWords = function()
	{
		session.ignore_words = "";
		var result = prefs.ignore_words.split(",");
		for(var i=0; i<result.length; i++)
		{
			var word = self.trim(result[i]);
			if(word == "")
				continue;
			
			if(session.ignore_words != "")
				session.ignore_words += "|";
			session.ignore_words += word;
		}
	};
	
	self.setUpBlockedUsers = function()
	{
		session.blocked_users = {};
		var result = prefs.blocked_users.split("\n");
		for(var i=0; i<result.length; i++)
		{
			session.blocked_users[self.trim(result[i])] = 1;
		}
	};
	
	
	self.addSnippet = function(target)
	{
		var snippets = elements[target + "_snippets"];
		var textarea = elements[target + "_contents"];
		
		if(!snippets || !textarea)
			return;
		
		var snippet = snippets.value;
		if(snippet == "")
			return;
		
		var start = textarea.selectionStart;
		var end = textarea.selectionEnd;
		if(snippet.charAt(0) == "\n")
		{
			if(start == 0 || (start > 0 && textarea.value.charAt(start - 1) == "\n"))
			{
				// At the head of line
				snippet = snippet.substr(1);
			}
			else if(textarea.value.charAt(start) == "\n")
			{
				// At the end of line
				start++;
				snippet = snippet.substr(1);
			}
			else
			{
				start = end = textarea.value.indexOf("\n", start);
				if(start < 0)
				{
					start = end = textarea.value.length;
				}
			}
		}
		
		textarea.value = textarea.value.substr(0, start) + snippet + textarea.value.substr(end);
		textarea.selectionStart = start + snippet.length;
	};
	
	
	self.resetSequence = function()
	{
		prefs.tweet_sequence = 0;
		session.test_tweet_sequence = 0;
		if(window.widget)
			widget.setPreferenceForKey(prefs.tweet_sequence, prefKeys.tweet_sequence);
		
		self.displayTweetSequence(prefs.tweet_sequence, session.tweet_contents);
		self.displayTestTweetSequence(session.test_tweet_sequence, elements.tweet_type2.checked? session.tweet_contents: null);
	};
	
	
	self.testTweet = function()
	{
		elements.tweet_test_result.innerHTML = "";
		
		var contents = self.parseContents("Tweets", elements.tweet_contents.value);
		if(contents)
		{
			var type = (elements.tweet_type1.checked? 1: (elements.tweet_type2.checked? 2: 0));
			var i, j, content, tweet;
			i = 0;
			while(++i < config.retry_count)
			{
				j = 0;
				while(++j < config.retry_count)
				{
					if(content = self.getTweet(contents, type, session.test_tweet_sequence, true))
						break;
				
					if(type == 2)
						break;
				}
				if(content)
				{
					tweet = self.evalContents(content, null, null, null, false, true);
					if(tweet && tweet != session.test_last_tweeted)
						break;
				}
			}
			if(!tweet)
				return;
			
			elements.tweet_test_result.innerHTML = tweet;
			session.test_last_tweeted = tweet;
			
			if(session.cleat_test_result_delay)
			{
				window.clearTimeout(session.cleat_test_result_delay);
				delete session.cleat_test_result_delay;
			}
			
			session.cleat_test_result_delay = window.setTimeout(
				function(e){ self.clearTestResult(); }, config.clear_test_result_interval * 1000);
			
			if(type == 2)
			{
				if(++session.test_tweet_sequence >= contents.children.length)
					session.test_tweet_sequence = 0;
			}
		}
		
		self.displayTestTweetSequence(session.test_tweet_sequence, elements.tweet_type2.checked? contents: null);
	};
	
	
	self.testScheduledTweet = function()
	{
		elements.scheduled_tweet_test_result.innerHTML = "";
		
		var contents = self.parseContents("Scheduled Tweets", elements.scheduled_tweet_contents.value, true);
		if(contents)
		{
			var date = self.constructDate({
				"year": elements.scheduled_tweet_test_year.value, 
				"month": elements.scheduled_tweet_test_month.value, 
				"day": elements.scheduled_tweet_test_day.value, 
				"hour": elements.scheduled_tweet_test_hour.value, 
				"minute": elements.scheduled_tweet_test_minute.value
			});
			
			var i, tweet;
			i = 0;
			while(++i < config.retry_count)
			{
				tweet = self.evalContents(contents, null, null, date, true, true);
				if(tweet != session.test_last_scheduled_tweeted)
					break;
			}
			
			elements.scheduled_tweet_test_result.innerHTML = tweet;
			session.test_last_scheduled_tweeted = tweet;
			
			if(session.cleat_test_result_delay)
			{
				window.clearTimeout(session.cleat_test_result_delay);
				delete session.cleat_test_result_delay;
			}
			
			session.cleat_test_result_delay = window.setTimeout(
				function(e){ self.clearTestResult(); }, config.clear_test_result_interval * 1000);
		}
	};
	
	
	self.testReply = function()
	{
		elements.reply_test_result.innerHTML = "";
		
		var contents = self.parseContents("Replies", elements.reply_contents.value);
		if(contents)
		{
			var user = session.user;
			var status = elements.reply_test.value.replace(/^(@\w+\s+)+/, "");
			if(session.ignore_words != "")
			{
				status = status.replace(new RegExp("(" + session.ignore_words + ").*"), "");
			}
			var reply = self.evalContents(contents, status, user, null, false, true);
			if(!reply)
				return;
			
			reply = reply.replace(config.replace_regexp, function(){ return self.replaceMatched(arguments, user); });
			reply = "@" + user.screen_name + " " + reply;
			
			elements.reply_test_result.innerHTML = reply;
			
			if(session.cleat_test_result_delay)
			{
				window.clearTimeout(session.cleat_test_result_delay);
				delete session.cleat_test_result_delay;
			}
			
			session.cleat_test_result_delay = window.setTimeout(
				function(e){ self.clearTestResult(); }, config.clear_test_result_interval * 1000);
		}
	};
	
	
	self.clearTestResult = function()
	{
		elements.tweet_test_result.innerHTML = "";
		elements.scheduled_tweet_test_result.innerHTML = "";
		elements.reply_test_result.innerHTML = "";
		session.cleat_test_result_delay = null;
	};
	
	
	//// Main
	self.startMainLoop = function()
	{
		if(config.debug)
			self.log(config.debug_log, "Main loop started.");
		
		if(session.main_loop_observer)
			window.clearInterval(session.main_loop_observer);
		
		prefs.active = true;
		if(window.widget)
			widget.setPreferenceForKey(prefs.active, prefKeys.active);
		
		elements.start_button.style.display = "none";
		elements.stop_button.style.display = "block";
		self.addClassName(elements.stop_button.parentNode, "active");
		self.startIndicatorBlinking();
		
		session.main_loop_observer = window.setInterval(function(){ self.mainLoop(); }, config.main_loop_interval * 1000);
		session.wait_for_request_mentions = 0;
		window.setTimeout(function(){ self.mainLoop(); }, 500);
	};
	
	
	self.stopMainLoop = function()
	{
		if(!prefs.active)
			return;
		
		if(config.debug)
			self.log(config.debug_log, "Main loop stopped.");
		
		prefs.active = false;
		if(window.widget)
			widget.setPreferenceForKey(prefs.active, prefKeys.active);
		
		elements.start_button.style.display = "block";
		elements.stop_button.style.display = "none";
		self.removeClassName(elements.stop_button.parentNode, "active");
		self.stopIndicatorBlinking();
		
		window.clearInterval(session.main_loop_observer);
		session.main_loop_observer = null;
	};
	
	
	self.startIndicatorBlinking = function()
	{
		elements.indicator.opacity = 0;
		elements.indicator.style.display = "block";
		self.turnOnIndicator();
	};
	
	
	self.stopIndicatorBlinking = function()
	{
		elements.indicator.style.display = "none";
		if(session.turn_on_indicator_animator)
		{
			session.turn_on_indicator_animator.stop();
			delete session.turn_on_indicator_animator;
			session.turn_on_indicator_animator = null;
		}
		if(session.turn_off_indicator_animator)
		{
			session.turn_off_indicator_animator.stop();
			delete session.turn_off_indicator_animator;
			session.turn_off_indicator_animator = null;
		}
	};
	
	
	self.mainLoop = function()
	{
		var now = (new Date).getTime();
		
		if(now >= prefs.last_tweeted_time + prefs.tweet_interval * 60 * 1000)
		{
			// Tweet
			window.setTimeout(function(){ self.tweet(); }, 0);
		}
		
		if(now >= prefs.last_scheduled_tweeted_time + 1 * 60 * 1000)
		{
			// Scheduled Tweet
			window.setTimeout(function(){ self.scheduledTweet(); }, 0);
		}
		
		if(prefs.sleep_until > 0)
		{
			// Wait until API limit reset
			if(now < prefs.sleep_until)
				return;
			
			prefs.sleep_until = 0;
			if(window.widget)
				widget.setPreferenceForKey(prefs.sleep_until, prefKeys.sleep_until);
		}
		
		if(--session.wait_for_request_mentions <= 0)
		{
			// Request mentions
			session.wait_for_request_mentions = prefs.request_mentions_interval * 60 / config.main_loop_interval;
			window.setTimeout(function(){ self.requestMentions(); }, 0);
		}
	};
	
	
	self.tweet = function()
	{
		if(session.tweeting)
			return;
		
		var i, j, content, tweet;
		i = 0;
		while(++i < config.retry_count)
		{
			j = 0;
			while(++j < config.retry_count)
			{
				if(content = self.getTweet(session.tweet_contents, prefs.tweet_type, prefs.tweet_sequence))
					break;
			
				if(prefs.tweet_type == 2)
					break;
			}
			
			if(content)
			{
				tweet = self.evalContents(content);
				if(tweet && tweet != prefs.last_tweeted)
					break;
			}
		}
		if(!tweet)
			return;
		
		// Tweet
		session.tweeting = true;
		
		prefs.last_tweeted_time = (new Date).getTime();
		prefs.last_tweeted = tweet;
		if(window.widget)
		{
			widget.setPreferenceForKey(prefs.last_tweeted_time, prefKeys.last_tweeted_time);
			widget.setPreferenceForKey(prefs.last_tweeted, prefKeys.last_tweeted);
		}
		
		tweet = tweet.substr(0, config.max_status_length);
		self.updateStatus(tweet, "");
	};
	
	
	self.tweetDidSucceed = function()
	{
		session.tweeting = false;
		
		if(prefs.tweet_type == 2)
		{
			if(++prefs.tweet_sequence >= session.tweet_contents.children.length)
				prefs.tweet_sequence = 0;
			
			if(window.widget)
				widget.setPreferenceForKey(prefs.tweet_sequence, prefKeys.tweet_sequence);
			
			self.displayTweetSequence(prefs.tweet_sequence, session.tweet_contents);
		}
	};
	
	
	self.tweetDidFailed = function()
	{
		session.tweeting = false;
	};
	
	
	self.scheduledTweet = function()
	{
		var tweet = self.evalContents(session.scheduled_tweet_contents, null, null, null, true);
		if(!tweet)
		{
			return;
		}
		
		// Tweet
		prefs.last_scheduled_tweeted_time = (new Date).getTime();
		if(window.widget)
			widget.setPreferenceForKey(prefs.last_scheduled_tweeted_time, prefKeys.last_scheduled_tweeted_time);
		
		tweet = tweet.substr(0, config.max_status_length);
		self.updateStatus(tweet, "");
	};
	
	
	self.checkMentions = function(mentions)
	{
		if(mentions.length < 1)
			return;
		
		var last_id = mentions[0].id_str;
		if(prefs.last_id == "")
		{
			// Skip old mentions
			for(var i=mentions.length-1; i>=0; i--)
			{
				var date = new Date(mentions[i].created_at);
				if(self.compareDate(date, session.launched_date) >= 0)
					break;
			}
			mentions.length = i + 1;
		}
		
		prefs.last_id = last_id;
		if(window.widget)
			widget.setPreferenceForKey(prefs.last_id, prefKeys.last_id);
		
		for(var i=0; i<mentions.length; i++)
		{
			if(!session.mentions_dict[mentions[i].id_str])
			{
				session.mentions.push(mentions[i]);
			}
		}
	};
	
	
	self.reply = function()
	{
		if(session.replying)
		{
			var mention = session.mentions.pop();
			delete session.mentions_dict[mention.id_str];
		}
		
		if(session.mentions.length < 1)
		{
			session.replying = false;
			return;
		}
		
		session.replying = true;
		
		var mention = session.mentions[session.mentions.length - 1];
		var user = mention.user;
		var status = mention.text.replace(/^(@\w+\s+)+/, "");
		if(session.ignore_words != "")
		{
			status = status.replace(new RegExp("(" + session.ignore_words + ").*"), "");
		}
		var date = new Date(mention.created_at);
		
		self.log("From @" + user.screen_name, mention.text, "from", date);
		
		if(user.screen_name == prefs.name || session.blocked_users[user.screen_name])
		{
			// Block reply from specifed user
			self.log(config.system_log, "Blocked @" + user.screen_name + ".");
			self.reply();
			return;
		}
		
		var reply = self.evalContents(session.reply_contents, status, user, date);
		if(!reply)
		{
			self.reply();
			return;
		}
		
		// Reply
		reply = reply.replace(config.replace_regexp, function(){ return self.replaceMatched(arguments, user); });
		reply = "@" + user.screen_name + " " + reply;
		
		reply = reply.substr(0, config.max_status_length);
		
		self.updateStatus(reply, mention.id_str);
	};
	
	
	self.replyDidSucceed = function()
	{
		self.reply();
	}
	
	
	self.replyDidFailed = function()
	{
		self.reply();
	}
	
	
	self.getTweet = function(contents, type, sequence, test)
	{
		if(!contents || !contents.children)
			return null;
		
		var content = null;
		if(type == 1)
		{
			// Random
			content = contents.children[Math.floor(Math.random() * contents.children.length)];
		}
		else if(type == 2)
		{
			// Sequential
			content = contents.children[sequence];
		}
		
		if(!content)
			return null;
		
		return self.evalContents({"root":true, "children":[content], "else_index":-1}, null, null, null, false, test);
	};
	
	
	self.evalContents = function(contents, mention, user, date, scheduled, test)
	{
		if(typeof(contents) == "string")
		{
			return contents;
		}
		
		var ret = "";
		
		if(!contents || !contents.children)
			return ret;
		
		var reply = (mention != null);
		var length = ((contents.else_index < 0)? contents.children.length: contents.else_index);
		for(var i=0; i<length; i++)
		{
			var condition = contents.children[i];
			var regexpMatched = null, dateMatched = null;
			
			if(typeof(condition) == "string")
			{
				ret = condition;
				break;
			}
			else if(reply && condition.type == "regexp")
			{
				if(!(regexpMatched = self.isMatchedRegExpCondition(condition, mention)))
					continue;
			}
			else if(reply && condition.type == "user" && user)
			{
				if(!self.isMatchedUserCondition(condition, user))
					continue;
			}
			else if(condition.type == "date")
			{
				var lastDate = ((!reply && !test)? (scheduled? prefs.last_scheduled_tweeted_time: prefs.last_tweeted_time): null);
				if(!self.isMatchedDateCondition(condition, date, lastDate))
					continue;
				
				dateMatched = self.deconstructDate(date? date: new Date);
			}
			else
			{
				continue;
			}
			
			// Matched
			if(typeof(condition.content) == "string")
			{
				ret = condition.content;
			}
			else
			{
				ret = self.evalContents(condition.content, mention, user, date, scheduled, test);
			}
			
			if(ret != "")
			{
				// Replace
				if(regexpMatched)
				{
					ret = ret.replace(config.replace_regexp, function(){ return self.replaceMatched(arguments, regexpMatched); });
				}
				if(dateMatched)
				{
					ret = ret.replace(config.replace_regexp, function(){ return self.replaceMatched(arguments, dateMatched); });
				}
				
				break;
			}
		}
		
		if(ret == "")
		{
			// Not matched
			if(contents.else_index >= 0)
			{
				ret = contents.children[
					contents.else_index + Math.floor(Math.random() * (contents.children.length - contents.else_index))
				];
			}
		}
		
		if(self.isUndefined(ret) || ret === false)
			ret = "";
		
		return ret;
	};
	
	
	self.replaceMatched = function(matched, values)
	{
		if(matched[1].match(/^\d+$/) || config.available_keys[matched[1]])
			return (!self.isUndefined(values[matched[1]])? values[matched[1]]: matched[0]);
		else
			return matched[0];
	};
	
	
	self.isMatchedRegExpCondition = function(condition, mention)
	{
		try
		{
			var regexp = new RegExp(condition.pattern, condition.options);
			return mention.match(regexp);
		}
		catch(e)
		{
			return null;
		}
	};
	
	
	self.isMatchedUserCondition = function(condition, user)
	{
		try
		{
			if(condition.operator != "=~")
			{
				return eval("(user." + condition.key + condition.operator + "\"" + condition.value + "\")");
			}
			else
			{
				var regexp = new RegExp(condition.regexp_pattern, condition.regexp_options);
				return user[condition.key].match(regexp);
			}
		}
		catch(e)
		{
			return null;
		}
	};
	
	
	self.isMatchedDateCondition = function(condition, date, lastDate)
	{
		if(!date)
			date = self.constructDate();
		
		if(!lastDate)
			lastDate = 0;
		
		var flag = false;
		for(var i=0; i<condition.spans.length; i++)
		{
			var span = condition.spans[i];
			if(span.length == 1)
			{
				// Exact condition
				var value = self.constructDate(span[0], date);
				if(self.compareDate(value, lastDate) > 0 && self.compareDate(value, date) == 0)
					flag = true;
			}
			else
			{
				// Span condition
				var value1 = self.constructDate(span[0], date);
				var value2 = self.constructDate(span[1], value1);
				if(self.compareDate(value1, lastDate) > 0 && self.compareDate(value1, date) <= 0 && self.compareDate(value2, date) >= 0)
					flag = true;
			}
		}
		
		if(flag)
		{
			if(condition.weekdays.length < 1)
				return true;
			
			// Check weekday condition
			var weekday = date.getDay();
			for(var i=0; i<condition.weekdays.length; i++)
			{
				if(config.weekdays[condition.weekdays[i]] == weekday)
					return true;
			}
		}
		
		return false;
	};
	
	
	//// Request to Twitter
	self.verifyCredentials = function()
	{
		session.last_verified_time = (new Date).getTime();
		
		if(window.dashbot_plugin)
		{
			self.showLoadingIndicator();
			dashbot_plugin.verifyCredentials();
		}
	};
	
	
	self.verifyCredentialsDidFinish = function(error, result)
	{
		self.hideLoadingIndicator();
		
		var response = null;
		if(result)
			response = self.evalResponse(result);
		
		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): L(config.err_failed_to_connect_to_twitter));
			self.displayError(errorString, true, true);
			return;
		}
		
		if(session.migrating_from)
		{
			// Remove previous preferences
			self.clearProfile();
			self.removePreferencesFor(session.migrating_from);
		}
		
		session.user = response;
		var name = response.screen_name;
		var image = response.profile_image_url;
		if(!image.match(/default_profile/))
			image = image.replace(/_normal/, "");
		
		if(name != prefs.name)
		{
			// Update recent accounts
			self.updateRecentAccounts(name, session.migrating_from);
		}
		
		self.setProfile(name, image);
		
		session.migrating_from = null;
		
		window.setTimeout(function(){ self.rateLimitStatus(); }, 0);
	}
	
	
	self.rateLimitStatus = function()
	{
		if(window.dashbot_plugin)
			dashbot_plugin.rateLimitStatus();
	};
	
	
	self.rateLimitStatusDidFinish = function(error, result)
	{
		var response = null;
		if(result)
			response = self.evalResponse(result);
		
		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): L(config.err_failed_to_connect_to_twitter));
			self.displayError(errorString, true, true);
			return;
		}
		
		self.setApiRemaining(response.remaining_hits, response.hourly_limit);
		
		if(response.remaining_hits < 1)
		{
			prefs.sleep_until = response.reset_time_in_seconds * 1000;
			if(window.widget)
				widget.setPreferenceForKey(prefs.sleep_until, prefKeys.sleep_until);
		}
	}
	
	
	self.requestMentions = function()
	{
		if(config.debug)
			self.log(config.debug_log, "Requesting mentions...");
		
		if(window.dashbot_plugin)
			dashbot_plugin.requestMentions(prefs.last_id);
	};
	
	
	self.requestMentionsDidFinish = function(error, result)
	{
		var response = null;
		if(result)
			response = self.evalResponse(result);
		
		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): L(config.err_failed_to_connect_to_twitter));
			self.displayError(errorString, true, true);
			return;
		}
		
		if(config.debug)
		{
			self.log(config.debug_log, "Requesting mentions completed - result = " + result);
		}
		
		window.setTimeout(function(){ self.rateLimitStatus(); }, 0);
		
		// Start replying
		self.checkMentions(self.evalResponse(result));
		self.reply();
	}
	
	
	self.updateStatus = function(status, status_id)
	{
		status = status.substr(0, config.max_status_length);
		if(window.dashbot_plugin)
			dashbot_plugin.updateStatus(status, status_id);
	};
	
	
	self.updateStatusDidFinish = function(error, result)
	{
		var response = null;
		if(result)
			response = self.evalResponse(result);
		
		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): L(config.err_failed_to_connect_to_twitter));
			self.displayError(errorString, true, true);
			if(session.tweeting)
			{
				self.tweetDidFailed();
			}
			else if(session.replying)
			{
				self.replyDidFailed();
			}
			return;
		}
		
		var log = response.text;
		log = log.replace(/&lt;/g, "<");
		log = log.replace(/&gt;/g, ">");
		log = log.replace(/&quot;/g, "\"");
		log = log.replace(/&amp;/g, "&");
		
		if(session.tweeting)
		{
			// Tweet
			self.log("", log, "tweet", new Date(response.created_at));
			self.tweetDidSucceed();
		}
		else if(session.replying)
		{
			// Reply
			self.log("To @" + response.user.screen_name, log, "to", new Date(response.created_at));
			self.replyDidSucceed();
		}
		else
		{
			// Scheduled tweet
			self.log("", log, "tweet", new Date(response.created_at));
		}
	};
	
	
	//// Other methods
	self.openProfile = function(username)
	{
		if(window.widget)
			widget.openURL("http://twitter.com/" + username);
	};
	
	
	self.migrateFrom = function(identifier)
	{
		if(!window.widget)
			return;
		
		if(!identifier)
			return;
		
		if(widget.preferenceForKey(identifier + "-oauth_token"))
		{
			// Relaunch
			session.migrating_from = identifier;
			window.setTimeout(function(){ self.launch(false); self.willShow(); }, 1000);
		}
	};
	
	
	//// For authorization
	self.requestRequestToken = function()
	{
		if(window.dashbot_plugin)
			dashbot_plugin.requestRequestToken();
	};
	
	
	self.requestRequestTokenDidFinish = function(error, result)
	{
		var response = null;
		if(result)
			response = self.evalResponse(result);
		
		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): "");
			self.displayError(L(config.err_failed_to_connect_to_twitter) + (errorString? ": " + errorString: ""), true);
			return;
		}
		
		elements.unauthorized.style.display = "none";
		elements.authorizing.style.display = "block";
		
		if(window.dashbot_plugin)
			dashbot_plugin.redirectToAuthorizeURL();
	}
	
	
	self.requestAccessToken = function()
	{
		var verifier = document.getElementById("verifier");
		if(window.dashbot_plugin)
			dashbot_plugin.requestAccessToken(verifier.value);
	};
	
	
	self.requestAccessTokenDidFinish = function(error, result)
	{
		var response = null;
		if(result)
			response = self.evalResponse(result);

		if(error != 0)
		{
			var errorString = (response? (response.error? response.error: response): "");
			self.displayError(L(config.err_failed_to_connect_to_twitter) + (errorString? ": " + errorString: ""), true);
			return;
		}
		
		elements.authorizing.style.display = "none";
		elements.main.style.display = "block";
		
		prefs.oauth_token = result;
		if(window.widget)
			widget.setPreferenceForKey(prefs.oauth_token, prefKeys.oauth_token);
		
		window.setTimeout(function(){ self.verifyCredentials(); }, 0);
	}
	
	
	//// Handle preferences
	self.selectTab = function(target)
	{
		if(!target)
			return;
		
		for(var i=0; i<elements.tabs.length; i++)
		{
			if(elements.tabs[i].id == target.id)
			{
				self.addClassName(elements.tabs[i], "active", "over");
				document.getElementById(elements.tabs[i].id + "_content").style.display = "block";
			}
			else
			{
				self.removeClassName(elements.tabs[i], "active", "over");
				document.getElementById(elements.tabs[i].id + "_content").style.display = "none";
			}
		}
		
		prefs.selected_tab = target.id;
		if(window.widget)
			widget.setPreferenceForKey(prefs.selected_tab, prefKeys.selected_tab);
	};
	
	
	//// Check for updates
	self.checkForUpdates = function()
	{
		prefs.last_checked_for_updates_time = (new Date).getTime();
		if(window.widget)
			widget.setPreferenceForKey(prefs.last_checked_for_updates_time, prefKeys.last_checked_for_updates_time);
		
		var request = new XMLHttpRequest();
		request.onreadystatechange = function(e){
			if(request.readyState == 4)
			{
				if(request.status != 200)
					return;
				
				var localVer = {ver:0, beta:0}, remoteVer = {ver:0, beta:0};
				
				if(config.version.match(/ver\.([\d.]+)(b(\d+))?/))
				{
					localVer.ver = RegExp.$1;
					localVer.beta = RegExp.$3;
				}
				
				if(request.responseText.match(/ver\.([\d.]+)(b(\d+))?/))
				{
					remoteVer.ver = RegExp.$1;
					remoteVer.beta = RegExp.$3;
				}
				
				if(
					(remoteVer.ver > localVer.ver) || 
					(remoteVer.ver == localVer.ver && ((!remoteVer.beta && localVer.beta) || (remoteVer.beta > localVer.beta)))
				)
				{
					// New version found
					document.getElementById("updated_button").style.display = "block";
				}
			}
		};
		request.open("GET", config.check_for_updates_url, true);
		request.setRequestHeader("Cache-Control", "no-cache");
		request.send("");
	}
	
	
	//// Event handlers
	self.willShow = function(e)
	{
		var now = (new Date).getTime();
		
		if(now >= session.last_checked_for_updates_time + config.check_for_updates_interval * 60 * 1000)
		{
			window.setTimeout(function(){ self.checkForUpdates(); }, 0);
		}
		
		if(!session.flipped && prefs.oauth_token != "" && dashbot_plugin.isAuthorized(prefs.oauth_token))
		{
			if(now >= session.last_verified_time + config.verification_interval * 60 * 1000)
			{
				window.setTimeout(function(){ self.verifyCredentials(); }, 750);
			}
		}
		
		if(!session.flipped && prefs.active && session.main_loop_observer)
		{
			self.startIndicatorBlinking();
		}
	};
	
	
	self.willHide = function(e)
	{
		if(!session.flipped)
			self.stopIndicatorBlinking();
	};
	
	
	self.didClickConnectButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.requestRequestToken();
	};
	
	
	self.didChangeRecentAccount = function(e)
	{
		if(session.error_shown)
			return;
		
		self.migrateFrom(e.target.value);
	};
	
	
	self.didClickEnterButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.requestAccessToken();
	};
	
	
	self.didClickCancelEnterButton = function(e)
	{
		if(session.error_shown)
			return;
		
		elements.unauthorized.style.display = "block";
		elements.authorizing.style.display = "none";
	};
	
	
	self.didClickStartButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.startMainLoop();
	};
	
	
	self.didClickStopButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.stopMainLoop();
	};
	
	
	self.didClickExpandButton = function(e)
	{
		self.expand();
	};
	
	
	self.didClickCollapseButton = function(e)
	{
		self.hideLog();
		// And then collapses
	};
	
	
	self.didClickImage = function(e)
	{
		self.openProfile(prefs.name);
	};
	
	
	self.didClickUpdatedButton = function(e)
	{
		if(window.widget)
			widget.openURL(config.support_url);
	};
	
	
	self.didClickInfoButton = function(e)
	{
		if(session.error_shown)
			return;
		
		session.active_status = prefs.active;
		
		self.stopMainLoop();
		window.setTimeout(function(){ self.showPreferences(); }, 0);
	};
	
	
	self.didMouseOverTab = function(e)
	{
		if(!self.hasClassName(e.target, "active"))
			self.addClassName(e.target, "over");
	};

	self.didMouseOutTab = function(e)
	{
		if(!self.hasClassName(e.target, "active"))
			self.removeClassName(e.target, "over");
	};
	
	
	self.didClickTab = function(e)
	{
		if(session.error_shown)
			return;
		
		self.selectTab(e.target);
	};
	
	
	self.didChangeTweetType = function(e)
	{
		if(elements.reset_sequence_button.disabled != elements.tweet_type1.checked)
			self.resetSequence();
		
		elements.reset_sequence_button.disabled = elements.tweet_type1.checked;
	}
	
	
	self.didClickTweetTestButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.testTweet();
	};
	
	
	self.didClickResetSequenceButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.resetSequence();
	};
	
	
	self.didClickScheduledTweetAddSnippetButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.addSnippet("scheduled_tweet");
	};
	
	
	self.didClickScheduledTweetTestButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.testScheduledTweet();
	};
	
	
	self.didClickReplyAddSnippetButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.addSnippet("reply");
	};
	
	
	self.didClickReplyTestButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.testReply();
	};
	
	
	self.didClickDoneButton = function(e)
	{
		if(session.error_shown)
			return;
		
		if(self.storePreferences())
		{
			self.savePreferences();
			self.hidePreferences();
			
			if(session.active_status)
			{
				window.setTimeout(function(){ self.startMainLoop(); }, 500);
			}
		}
	};
	
	
	self.didClickCancelButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.restorePreferences();
		self.hidePreferences();
		
		if(session.active_status)
		{
			window.setTimeout(function(){ self.startMainLoop(); }, 500);
		}
	};
	
	
	self.didClickClearLogButton = function(e)
	{
		if(session.error_shown)
			return;
		
		self.clearLog();
		self.displayError(L("Log has been cleared."));
	};
	
	
	self.didClickOkButton = function(e)
	{
		self.hideError();
	};
	
	
	//// Utility methods
	self.isUndefined = function(obj)
	{
		return (typeof(obj) == "undefined");
	};
	
	
	self.hasClassName = function(obj, name)
	{
		var regexp = new RegExp("\\b" + name + "\\b");
		return obj.className.match(regexp);
	};
	
	
	self.addClassName = function(obj)
	{
		for(var i=1; i<arguments.length; i++)
		{
			if(!self.hasClassName(obj, arguments[i]))
				obj.className += ((obj.className != "")? " ": "") + arguments[i];
		}
	};
	
	
	self.removeClassName = function(obj)
	{
		for(var i=1; i<arguments.length; i++)
		{
			var regexp = new RegExp("^" + arguments[i] + "\\s*|\\s*" + arguments[i] + "\\b", "g");
			obj.className = obj.className.replace(regexp, "");
		}
	};


	self.evalResponse = function(response)
	{
		try
		{
			return eval("(" + response + ")");
		}
		catch(e)
		{
			return self.evalResponse("\"" + response + "\"");
		}
	};
	
	
	self.dumpObject = function(obj)
	{
		var dump = "";
		for(i in obj)
		{
			dump += i + " : " + obj + "\n";
		}
		return dump;
	};
	
	
	self.zeroFill = function(num, digit)
	{
		for(var i=0; i<digit; i++)
		{
			num = "0" + num;
		}
		
		return num.substr(num.length - digit, digit);
	};
	
	
	self.trim = function(text)
	{
		return text.replace(/^\s*(.+?)\s*$/, "$1");
	};
	
	
	self.stringFromMinutes = function(minutes)
	{
		var value = minutes;
		var suffix = "";
		if(value >= 60 && value % 60 == 0)
		{
			value = value / 60;
			suffix = L((value > 1)? "hours": "hour");
		}
		else
		{
			suffix = L((value > 1)? "minutes": "minute");
		}
		return value + " " + suffix;
	};
	
	
	self.constructDate = function(source, base)
	{
		if(!base)
		{
			base = new Date;
			base.setSeconds(0);
			base.setMilliseconds(0);
		}
		
		if(!source || source == "*")
			return base;
		
		var baseDate = self.deconstructDate(base);
		var ret = new Date(
			(source.year   == "*" || source.year   == "")? baseDate.year     : source.year, 
			(source.month  == "*" || source.month  == "")? baseDate.month - 1: source.month - 1, 
			(source.day    == "*" || source.day    == "")? baseDate.day      : source.day, 
			(source.hour   == "*" || source.hour   == "")? baseDate.hour     : source.hour, 
			(source.minute == "*" || source.minute == "")? baseDate.minute   : source.minute, 
			baseDate.second, 
			0
		);
		return ret;
	};
	
	
	self.deconstructDate = function(date)
	{
		var year = date.getYear(); if(year < 1900) year += 1900;
		var month = date.getMonth() + 1;
		var day = date.getDate();
		var hour = date.getHours();
		var minute = date.getMinutes();
		var second = date.getSeconds();
		
		return {"year": year, "month": month, "day": day, "hour": hour, "minute": minute, "second": second};
	}
	
	
	self.compareDate = function(date1, date2)
	{
		var t1 = ((typeof(date1) == "object")? date1.getTime(): date1);
		var t2 = ((typeof(date2) == "object")? date2.getTime(): date2);
		if(t1 < t2)
			return -1;
		else if(t1 > t2)
			return 1;
		else
			return 0;
	};
	
	
	self.initialize();
};


//////// Callbacks
function requestRequestTokenDidFinish(error)
{
	if(window.controller)
		window.controller.requestRequestTokenDidFinish(error);
}


function requestAccessTokenDidFinish(error, result)
{
	if(window.controller)
		window.controller.requestAccessTokenDidFinish(error, result);
}


function verifyCredentialsDidFinish(error, result)
{
	if(window.controller)
		window.controller.verifyCredentialsDidFinish(error, result);
}


function rateLimitStatusDidFinish(error, result)
{
	if(window.controller)
		window.controller.rateLimitStatusDidFinish(error, result);
}


function requestMentionsDidFinish(error, result)
{
	if(window.controller)
		window.controller.requestMentionsDidFinish(error, result);
}


function updateStatusDidFinish(error, result)
{
	if(window.controller)
		window.controller.updateStatusDidFinish(error, result);
}


//// Get localized string
function L(text)
{
	if(window.dashcode)
		return dashcode.getLocalizedString(text);
	else
		return text;
}


document.addEventListener("DOMContentLoaded", function(e){ window.controller = new DashbotController; });
